var changesMade = false;
var entryChangesMade = false;  // used to keep track of unsaved changes from one screen to the next.
var dataObj;
var dontPromptWhenLeaving = false;  // used by the MainMenu and Logout links to inform the onbeforeunload handler to not prompt.

// This holds the info needed to try and automate the display of info on the screen.
var nameLookups = new Array;

// If the result is false, don't continue setting up a new screen, else continue.
function verifyChangesSaved()
{
  if (entryChangesMade)
  {
    var result = confirm("You have uncommitted changes!\nHit Cancel to return and Update,\n or OK to continue and lose the changes.");
    return result;
  }
  return true;
}

// Returns the Uncomitted Changes string if changes have been made, else an empty string to indicate we don't need to ask the user.
function verifyChangesSavedWindowClosing()
{
  if (changesMade || entryChangesMade)
  {
    if (dontPromptWhenLeaving)
    {
      return;
    }
    else
    {
      return "You have uncommitted changes!";
    }
  }
  else if (dontPromptWhenLeaving)
  {
    return;
  }
  return "";  // prompt in all other situations so that we can minimize the accidental closing of the window/browser or using the back button.
}

function triggerForm(command)
{
  var d = document.editF;

  if (command == "mainMenu")
  {
    if (changesMade || entryChangesMade)
    {
      var result = confirm("Are you sure you want to return to the Main Menu without saving?");
      if (result == false)
      {
        return;
      }
    }
    dontPromptWhenLeaving = true;
  }
  if (command == "logout")
  {
    if (changesMade || entryChangesMade)
    {
      var result = confirm("Are you sure you want to Logout without saving?");
      if (result == false)
      {
        return;
      }
    }
    dontPromptWhenLeaving = true;
  }
  d.command.value = command;
  d.submit();
}

function convertConfigToXML()
{
  // walk the dataObj and generate the XML document to return to the
  // backend for saving.

  var output = '<?xml version="1.0" encoding="ISO-8859-1"?>\n';
  output += '<firewall module="' + dataObj.info.module + '" moduleVersion="' + dataObj.info.version + '">\n';

  // output <config> info
  var configObj = dataObj.config;
  output += '  <config>\n';
  output += '    <networkCommand command="' + configObj.networkCommand + '"/>\n';
  output += '    <mangleRules enabled="' + (configObj.mangleRules ? 'true' : 'false') + '"/>\n';
  output += '    <ecn enabled="' + (configObj.ecn ? 'true' : 'false') + '"/>\n';
  output += '    <bridgeSupport enabled="' + (configObj.bridgeSupport ? 'true' : 'false') + '"/>\n';
  output += '    <snortInlineSupport enabled="' + (configObj.snortInlineSupport ? 'true' : 'false') + '"/>\n';
  output += '    <dynamicInterfaces mode="' + configObj.dynamicInterfacesMode + '"/>\n';
  output += '    <validityChecks tcpFlags="' + (configObj.validityChecks['tcpFlags'] ? 'true' : 'false') + '" unclean="' + (configObj.validityChecks['unclean'] ? 'true' : 'false') + '"/>\n';
  output += '    <moduleSupport kernel="' + (configObj.moduleSupport.modular ? 'modular' : 'monolithic') + '">\n';
  for (var i in configObj.moduleSupport.conntrackers)
  {
    output += '      <conntracker name="' + i + '" params="' + configObj.moduleSupport.conntrackers[i].params + '" active="' + (configObj.moduleSupport.conntrackers[i].active ? 'true' : 'false') + '"/>\n';
  }
  for (var i in configObj.moduleSupport.modules)
  {
    output += '      <module name="' + i + '"/>\n';
  }
  output += '    </moduleSupport>\n';
  for (var i in configObj.zones)
  {
    output += '    <' + i +' enabled="' + (configObj.zones[i] ? 'true' : 'false') + '"/>\n';
  }
  output += '    <log prefix="' + configObj.logInfo.prefix + '" level="' + configObj.logInfo.level + '"/>\n';
  output += '    <logRate';
  for (var i in configObj.limitRates['log'])
  {
    output += ' ' + i + '="' + configObj.limitRates['log'][i] + '"';
  }
  output += '/>\n';
  output += '    <limitRate icmp="' + configObj.limitRates['icmp'] + '"/>\n';
  output += '  </config>\n';

  // output the <networks> info.
  var networks = dataObj.networks;
  output += '  <networks>\n';
  for (var i in networks)
  {
    output += '    <addr value="' + networks[i].host + '" type="' + networks[i].type + '" index="' + i + '" active="' + (networks[i].active ? 'true' : 'false') + '" comment="' + networks[i].comment + '"/>\n';
  }
  output += '  </networks>\n';

  // output the <reserved> info.
  var reserved = dataObj.reserved;
  output += '  <reserved>\n';
  for (var i in reserved)
  {
    var entry = reserved[i];
    output += '    <addr value="' + entry.host + '" active="' + (entry.active ? 'true' : 'false') + '" limitTo="';
    output += entry.limitTo.join();
    output += '" comment="' + entry.comment + '"/>\n';
  }
  output += '  </reserved>\n';

  // output the <zones> info
  output += '  <zones>\n';
  var zones = dataObj.zones;
  for (var zone in zones)
  {
    if (zone == "aliasLookup")
    {
      continue;
    }
    if (numHashEntries(zones[zone]) > 0)
    {
      output += '    <zone type="' + zone + '">\n';
      for (var alias in zones[zone])
      {
        var info = zones[zone][alias];
        
        // make sure this interface is being used, if it is active.
        if (info.active && (configObj.zones[zone] || zone == 'external'))
        {
          // find out how many paths this interface is used in and count *.
          var namedPaths = findInUsePaths('interface', alias);
          var starPaths = findInUsePaths('zone', zone); // handles the * case.
          
          if (namedPaths + starPaths == 0)
          {
            throw new Error("convertConfigToXML():  alias '" + alias + "' in zone '" + zone + "' is enabled but is not referenced in any paths!");
          }
        }
        
        output += '      <interface name="' + info.name + '" alias="' + alias + '"';
        if (zone != 'ipsec')
        {
          output += ' ip="' + info.ip + '" ipsec="' + (info.ipsec ? 'true' : 'false') + '" ipsecLimitFrom="' + info.ipsecLimitFrom.join() + '"';
        }
        else
        {
          output += ' boundTo="' + info.boundTo + '"';
        }
        output += ' network="' + info.network.join() + '"';
        output += ' proxyarp="' + (info.proxyarp ? 'true' : 'false') + '" active="' + (info.active ? 'true' : 'false') + '" comment="' + info.comment + '"/>\n';
      }
      output += '    </zone>\n';
    }
    else
    {
      if (configObj.zones[zone] || zone == 'external')
      {
        throw new Error("convertConfigToXML():  zone '" + zone + "' is enabled but has no interfaces defined!");
      }
    }
  }
  output += '  </zones>\n';

  // output the <services> info
  output += '  <services>\n';
  var services = dataObj.services;
  for (var serviceIndex in services['servicesOrder'])
  {
    var service = services['servicesOrder'][serviceIndex];
    output += '    <service name="' + service + '" comment="' + services[service].comment + '">\n';
    for (var infoIndex in services[service].infos)
    {
      var info = services[service].infos[infoIndex];
      output += '      <info proto="' + info.proto + '"';
      if (info.proto == 'icmp')
      {
        output += ' icmpType="' + info.icmpType + '"';
      }

      {
        output += ' dport="' + info.dport + '" sport="' + info.sport + '"';
      }
      output += '/>\n';
    }
    output += '    </service>\n';
  }
  output += '  </services>\n';

  // output the <paths> info
  output += '  <paths>\n';
  var paths = dataObj.paths;
  for (var serviceIndex in paths['serviceGroupsOrder'])
  {
    var serviceGroup = paths['serviceGroupsOrder'][serviceIndex];
    output += '    <serviceGroup name="' + serviceGroup + '" active="' + (paths[serviceGroup]['active'] ? 'true' : 'false') + '" comment="' + paths[serviceGroup]['comment'] + '">\n';
    for (var pathIndex in paths[serviceGroup]['zonePathOrder'])
    {
      var path = paths[serviceGroup]['zonePathOrder'][pathIndex];
      for (var infoIndex in paths[serviceGroup][path])
      {
        var info = paths[serviceGroup][path][infoIndex];
        output += '      <path type="' + path + '"';
        if (info.i.length > 0)
        {
          output += ' i="' + info.i.join() + '"';
        }
        if (info.o.length > 0)
        {
          output += ' o="' + info.o.join() + '"';
        }
        if (info.ip.length > 0)
        {
          output += ' ip="' + info.ip + '"';
        }
        if (info.limitFrom.length > 0)
        {
          output += ' limitFrom="' + info.limitFrom.join() + '"';
        }
        if (info.limitTo.length > 0)
        {
          output += ' limitTo="' + info.limitTo.join() + '"';
        }
        output += ' active="' + (info.active ? 'true' : 'false') + '" comment="' + info.comment + '">\n';
        if (info.reject)
        {
          output += '        <reject/>\n';
        }
        if (info.dropNoLog)
        {
          output += '        <dropNoLog/>\n';
        }
        if (info.dnat)
        {
          output += '        <dnat destination="' + info.dnatDestination + '" zone="' + info.dnatZone + '" interface="' + info.dnatInterface + '" dnatPort="' + info.dnatPort + '"/>\n';
        }
        if (info.redirect)
        {
          output += '        <redirect redirectPort="' + info.redirectPort + '"/>\n';
        }
        if (info.queue)
        {
          output += '        <queue/>\n';
        }
        if (info.log)
        {
          output += '        <log comment="' + info.logComment + '" level="' + info.logLevel + '"/>\n';
        }
        for (var serviceEntryIndex in info.service)
        {
          output += '        <service name="' + info.service[serviceEntryIndex] + '"/>\n';
        }

        output += '      </path>\n';
      }
    }
    output += '    </serviceGroup>\n';
  }
  output += '  </paths>\n';

  output += '</firewall>\n';

  return output;
}

function getContent()
{
  return document.getElementById("content");
}

function getStatus()
{
  return document.getElementById("status");
}

function getMenuBar()
{
  return document.getElementById("menubar");
}

function displayXML()
{
  if (! verifyChangesSaved())
  {
    return;
  }
  entryChangesMade = false;

  try
  {
    var xml = convertConfigToXML();
  }
  catch (e)
  {
    displayStatus(e.message, 'error');

    // get our content section
    var content = document.getElementById("content");

    // replace the body of the content block.
    content.innerHTML = "";
    return;
  }
  xml = xml.replace(/&/g, "&amp;");
  xml = xml.replace(/</g, "&lt;");
  xml = xml.replace(/>/g, "&gt;");
  xml = xml.replace(/  /g, "&nbsp; ");
  xml = xml.replace(/\n/gm, "<br>\n");

  displayStatus('Here is the current XML config file that would be sent to the backend for saving...', 'info');

  // get our content section
  var content = document.getElementById("content");

  // replace the body of the content block.
  content.innerHTML = "<b>XML</b><br><code>" + xml + "</code>";
}

// return the number of valid zones defined.
function numValidZones()
{
  var zones = dataObj.zones;
  var counter = 0;
  for (var zone in zones)
  {
    if (zone == "aliasLookup")
    {
      continue;
    }
    counter += numHashEntries(zones[zone]);
  }
  return counter;
}

function saveConfig()
{
  if (! verifyChangesSaved())
  {
    return false;
  }
  entryChangesMade = false;

  try
  {
    var xml = convertConfigToXML();
  }
  catch (e)
  {
    displayStatus(e.message, 'error');

    // get our content section
    var content = document.getElementById("content");

    // replace the body of the content block.
    content.innerHTML = "";
    return false;
  }

  // make sure we appear to have a fully valid scenario (zones, services and serviceGroups not-empty)
  if (numHashEntries(dataObj.networks) == 0 || numValidZones() == 0 || numHashEntries(dataObj.services['servicesOrder']) == 0 || numHashEntries(dataObj.paths['serviceGroupsOrder']) == 0)
  {
    displayXML();
    displayStatus('You have an invalid XML config file in that one or more of your Networks, Zones, Services or ServiceGroups are empty!', 'error');

    return false;
  }

  document.editF.extraArgs.value = xml;
  dontPromptWhenLeaving = true;
  
  return true;
}

function processReset()
{
  if (changesMade || entryChangesMade)
  {
    var result = confirm("Are you sure you want to Reset to initial values?");
    if (result == false)
    {
      return;
    }
  }
  changesMade = false;  // make sure the flag is turned off!
  entryChangesMade = false;
  resetConfig();
  redisplayEditScreen();
}

// name: findInUsePaths
// returns: number of paths found to be using this entry.
// takes: type - service, interface, network, zone
//        value - the service, interface or network to find in the paths.
//              if type = zone, then value is the name of the zone that
//              we should check to make sure paths to/from that zone have
//              the * interface defined.
// purpose: to find the specified type in the paths that reference it and
//          return the count.
function findInUsePaths(type, value)
{
  var pathCount = 0;
  if (type == 'service' || type == 'interface' || type == 'network' || type == 'zone')
  {
    var paths = dataObj.paths;
    for (var serviceIndex in paths['serviceGroupsOrder'])
    {
      var serviceGroup = paths['serviceGroupsOrder'][serviceIndex];
      for (var serviceIndex in paths[serviceGroup]['zonePathOrder'])
      {
        var path = paths[serviceGroup]['zonePathOrder'][serviceIndex];
        for (var infoIndex in paths[serviceGroup][path])
        {
          var info = paths[serviceGroup][path][infoIndex];

          if (type == 'service')
          {
            for (var servicesIndex in info.service)
            {
              if (info.service[servicesIndex] == value)
              {
                pathCount++;
                break;
              }
            }
          }
          else if (type == 'interface')
          {
            if (info.i.length > 0)
            {
              for (var aliasIndex in info.i)
              {
                if (info.i[aliasIndex] == value)
                {
                  pathCount++;
                  break;
                }
              }
            }
            if (info.o.length > 0)
            {
              for (var aliasIndex in info.o)
              {
                if (info.o[aliasIndex] == value)
                {
                  pathCount++;
                  break;
                }
              }
            }
          }
          else if (type == 'network')
          {
            var limitTo = info.limitTo;
            var limitFrom = info.limitFrom;
            var ip = info.ip;

            for (var hostIndex in limitTo)
            {
              if (limitTo[hostIndex] == value)
              {
                pathCount++;
                break;
              }
            }
            for (var hostIndex in limitFrom)
            {
              if (limitFrom[hostIndex] == value)
              {
                pathCount++;
                break;
              }
            }
            if (ip == value)
            {
              pathCount++;
            }
            // check for dnat with an destination = value.
            if (info.dnat && info.dnatDestination == value)
            {
              pathCount++;
            }
          }
          else if (type == 'zone')
          {
            var normalPattern = /^(.+)To(.+)$/;
            var broadcastPattern = /^(.+)Broadcast$/;
      
            var result = path.match(normalPattern);
            var from;
            var to;
            var found = false;
            if (result != null)
            {
              from = result[1].toLowerCase();  // The from zone
              to = result[2].toLowerCase();  // The to zone
              
              if (from == value || to == value)
              {
                found = true;
              }
            }
            else
            {
              result = path.match(broadcastPattern);
              if (result != null)
              {
                from = result[1].toLowerCase();  // The from zone
                if (from == value)
                {
                  found = true;
                }
              }
            }
            
            if (found)
            {
              // now we loop over the info looking for * entries.
              if (info.i.length > 0)
              {
                for (var aliasIndex in info.i)
                {
                  if (info.i[aliasIndex] == '*')
                  {
                    pathCount++;
                    break;
                  }
                }
              }
              if (info.o.length > 0)
              {
                for (var aliasIndex in info.o)
                {
                  if (info.o[aliasIndex] == '*')
                  {
                    pathCount++;
                    break;
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  else
  {
    alert("Invalid type specified!  type='" + type + "'");
    return 0;
  }
  return pathCount;
}

// name: findBoundToInterfaces
// takes: zone, alias
// returns: number of interfaces that are boundTo us in the ipsec zone, as long as zone != ipsec.
function findBoundToInterfaces(zone, alias)
{
  var boundToCount = 0;
  var zones = dataObj.zones;
  if (zone != 'ipsec')
  {
    // make sure we are not referenced by any ipsec interfaces.
    for (var i in zones['ipsec'])
    {
      if (zones['ipsec'][i].boundTo == alias)
      {
        boundToCount++;
      }
    }
  }
  return boundToCount;
}

// name: clearBoundToInterface
// takes: alias
// returns: true on success, false on error
// sets the ipsec flag to false for the interface that alias was boundTo.
// also empties the ipsecLimitFrom entry.
function clearBoundToInterface(alias)
{
  var zones = dataObj.zones;

  // only do the ipsec disabling if there is only 1 ipsec interface boundTo alias.
  if (findBoundToInterfaces('external', alias) > 1)
  {
    return true;
  }
  for (var zone in zones)
  {
    if (numHashEntries(zones[zone]) > 0 && zone != 'ipsec' && zone != 'aliasLookup')
    {
      for (var i in zones[zone])
      {
        if (i == alias)
        {
          zones[zone][i].ipsec = false;
          zones[zone][i].ipsecLimitFrom.length = 0;
          return true;
        }
      }
    }
  }
  alert("Could not find interface '" + alias + "' to clear it's ipsec flag!");
  return false;
}

// name: clearAllBoundToInterfaces
// takes: nothing
// returns: nothing
// sets the ipsec flag to false on all non-ipsec interfaces.
function clearAllBoundToInterfaces()
{
  var zones = dataObj.zones;

  for (var zone in zones)
  {
    if (numHashEntries(zones[zone]) > 0 && zone != 'ipsec' && zone != 'aliasLookup')
    {
      for (var alias in zones[zone])
      {
        zones[zone][alias].ipsec = false;
        zones[zone][alias].ipsecLimitFrom.length = 0;
      }
    }
  }
}

// name: findLimitToReservedEntries
// takes: alias
// returns: number of Reserved Entries that have us in their limitTo array.
function findLimitToReservedEntries(alias)
{
  var limitToCount = 0;
  var reserved = dataObj.reserved;
  
  for (var i in reserved)
  {
    for (var j in reserved[i].limitTo)
    {
      if (reserved[i].limitTo[j] == alias)
      {
        limitToCount++;
      }
    }
  }
  
  return limitToCount;
}

// name: deleteReservedLimitToEntry
// returns: boolean (true = all ok, false = error occured)
// takes: alias - interface name to remove from the limitTo arrays.
function deleteReservedLimitToEntry(alias)
{
  var reserved = dataObj.reserved;
  
  for (var i in reserved)
  {
    for (var j in reserved[i].limitTo)
    {
      if (reserved[i].limitTo[j] == alias)
      {
        // now to delete it.
        reserved[i].limitTo.splice(j, 1);
        changesMade = true;
        break;
      }
    }
  }
  
  return true;
}

// name: updateReservedLimitToEntries
// returns: boolean (true = all ok, false = error occured)
// takes: alias - interface name to update in the limitTo arrays.
//        newAlias - the new interface name to change alias to.
function updateReservedLimitToEntries(alias, newAlias)
{
  var reserved = dataObj.reserved;
  
  for (var i in reserved)
  {
    for (var j in reserved[i].limitTo)
    {
      if (reserved[i].limitTo[j] == alias)
      {
        // now update it.
        reserved[i].limitTo[j] = newAlias;
        changesMade = true;
        break;
      }
    }
  }
  
  return true;
}

// name: deleteInUsePaths
// returns: boolean (true = all ok, false = error occured)
// takes: type - service, interface, network, zone
//        value - the service, interface or network to remove from the paths.
//              if type = zone, then value is the name of the zone that
//              we should check to make sure paths to/from that zone have
//              the * interface removed.
// purpose: to remove the specified type from the paths that reference it, possibly
//          removing the paths and serviceGroups as they are emptied.
// First, find the InUsePaths and if they are > 0, ask the user to delete.
// If the user says no, return false, to indicate the caller should not proceed,
// Else, proceed to delete the entries that are being used in the paths.
function deleteInUsePaths(type, value)
{
  if (type == 'service' || type == 'interface' || type == 'network' || type == 'zone')
  {
    var numPaths = findInUsePaths(type, value);

    if (numPaths > 0)
    {
      var result = confirm("Would you like to delete the " + numPaths + " paths that reference " + type + " '" + value + "'?\nClick Ok to Delete them, Cancel to abort the delete " + type + " request.");
      if (result == false)
      {
        return false;
      }
      // now we loop back over the paths and delete the references to this type (service, interface, network, * interfaces).
      var paths = dataObj.paths;
      var serviceGroupsToDelete = new Array();
      for (var serviceIndex in paths['serviceGroupsOrder'])
      {
        var serviceGroup = paths['serviceGroupsOrder'][serviceIndex];
        var zonesToDelete = new Array();
        for (var zoneIndex in paths[serviceGroup]['zonePathOrder'])
        {
          var path = paths[serviceGroup]['zonePathOrder'][zoneIndex];
          var pathsToDelete = new Array();
          for (var infoIndex in paths[serviceGroup][path])
          {
            var info = paths[serviceGroup][path][infoIndex];

            if (type == 'service')
            {
              for (var servicesIndex in info.service)
              {
                if (info.service[servicesIndex] == value)
                {
                  // delete this index entry.
                  info.service.splice(servicesIndex, 1);
                  changesMade = true;
                  break;
                }
              }
              if (info.service.length == 0)
              {
                // we just emptied the path, so delete it.
                pathsToDelete.push(infoIndex);
              }
            }
            else if (type == 'interface')
            {
              if (info.i.length > 0)
              {
                for (var aliasIndex in info.i)
                {
                  if (info.i[aliasIndex] == value)
                  {
                    // delete this index entry.
                    info.i.splice(aliasIndex, 1);
                    changesMade = true;
                    break;
                  }
                }
                if (info.i.length == 0)
                {
                  // we just emptied the path, so delete it.
                  pathsToDelete.push(infoIndex);
                }
              }
              if (info.o.length > 0 && pathsToDelete[pathsToDelete.length-1] != infoIndex)
              {
                for (var aliasIndex in info.o)
                {
                  if (info.o[aliasIndex] == value)
                  {
                    // delete this index entry.
                    info.o.splice(aliasIndex, 1);
                    changesMade = true;
                    break;
                  }
                }
                if (info.o.length == 0)
                {
                  // we just emptied the path, so delete it.
                  pathsToDelete.push(infoIndex);
                }
              }
              // check for dnat with an interface = value.
              if (info.dnat && info.dnatInterface == value)
              {
                pathsToDelete.push(infoIndex);
              }
            }
            else if (type == 'network')
            {
              var limitTo = info.limitTo;
              var limitFrom = info.limitFrom;
              var ip = info.ip;

              if (limitTo.length > 0)
              {
                for (var hostIndex in limitTo)
                {
                  if (limitTo[hostIndex] == value)
                  {
                    limitTo.splice(hostIndex, 1);
                    changesMade = true;
                    break;
                  }
                }
                if (limitTo.length == 0)
                {
                  // we just emptied the path, so delete it.
                  pathsToDelete.push(infoIndex);
                }
              }
              if (limitFrom.length > 0)
              {
                for (var hostIndex in limitFrom)
                {
                  if (limitFrom[hostIndex] == value)
                  {
                    limitFrom.splice(hostIndex, 1);
                    changesMade = true;
                    break;
                  }
                }
                if (limitFrom.length == 0)
                {
                  // we just emptied the path, so delete it.
                  pathsToDelete.push(infoIndex);
                }
              }
              if (ip == value)
              {
                ip = "";  // This may break this path, so do we force it to be deleted?
              }
              // check for dnat with an destination = value.
              if (info.dnat && info.dnatDestination == value)
              {
                pathsToDelete.push(infoIndex);
              }
            }
            else if (type == 'zone')
            {
              var normalPattern = /^(.+)To(.+)$/;
              var broadcastPattern = /^(.+)Broadcast$/;
        
              var result = path.match(normalPattern);
              var from;
              var to;
              var found = false;
              if (result != null)
              {
                from = result[1].toLowerCase();  // The from zone
                to = result[2].toLowerCase();  // The to zone
                
                if (from == value || to == value)
                {
                  found = true;
                }
              }
              else
              {
                result = path.match(broadcastPattern);
                if (result != null)
                {
                  from = result[1].toLowerCase();  // The from zone
                  if (from == value)
                  {
                    found = true;
                  }
                }
              }
              
              if (found)
              {
                // now we loop over the info looking for * entries.
                // if we find one, then we are deleting this path as the i/o can only contain * or non-* entries.
                if (info.i.length > 0)
                {
                  for (var aliasIndex in info.i)
                  {
                    if (info.i[aliasIndex] == '*')
                    {
                      pathsToDelete.push(infoIndex);
                      break;
                    }
                  }
                }
                if (info.o.length > 0)
                {
                  for (var aliasIndex in info.o)
                  {
                    if (info.o[aliasIndex] == '*')
                    {
                      pathsToDelete.push(infoIndex);
                      break;
                    }
                  }
                }
              }
            }
          }
          // delete any paths that were marked for deletion.
          for (var pathIndex=0; pathIndex < pathsToDelete.length; pathIndex++)
          {
            var index = pathsToDelete[pathIndex];
            index -= pathIndex;  // adjust for the shrinking array size.
            paths[serviceGroup][path].splice(index, 1);
            changesMade = true;
          }
          // make sure we just didn't empty this serviceGroup path entry
          if (numHashEntries(paths[serviceGroup][path]) == 0)
          {
            delete paths[serviceGroup][path];
            changesMade = true;
            zonesToDelete.push(zoneIndex);
          }
        }
        // delete any zones that were marked for deletion.
        for (var zoneIndex=0; zoneIndex < zonesToDelete.length; zoneIndex++)
        {
          var index = zonesToDelete[zoneIndex];
          index -= zoneIndex;  // adjust for the shrinking array size.
          paths[serviceGroup]['zonePathOrder'].splice(index, 1);
          changesMade = true;
        }
        // make sure we just didn't empty this serviceGroup
        if (paths[serviceGroup]['zonePathOrder'].length == 0)
        {
          delete paths[serviceGroup];
          changesMade = true;
          serviceGroupsToDelete.push(serviceIndex);
        }
      }
      // delete any serviceGroups that were marked for deletion.
      for (var serviceIndex=0; serviceIndex < serviceGroupsToDelete.length; serviceIndex++)
      {
        var index = serviceGroupsToDelete[serviceIndex];
        index -= serviceIndex;  // adjust for the shrinking array size.
        paths['serviceGroupsOrder'].splice(index, 1);
        changesMade = true;
      }
    }
  }
  else
  {
    alert("Invalid type specified!  type='" + type + "'");
    return false;
  }
  return true;
}

function redisplayEditScreen()
{
  var status = getStatus();
  var content = getContent();

  if (! verifyChangesSaved())
  {
    return;
  }
  entryChangesMade = false;

  displayStatus("Ready...", "info");

  var html = 'Welcome to the JavaScript Editor for <a href="http://pcxfirewall.sf.net/" target="links">PCXFirewall</a> VeryTight2 config files.<br><br>\n';
  html += 'You probably want to go about editing your firewall scenario as follows:<br><br>\n';
  html += '<ol>\n';
  html += '  <li><a href="#" onClick="displayConfig(); return false;" title="Edit &lt;config&gt; section">Config</a> - Specify the global settings for your firewall.  You enable extra Zones here.</li>\n';
  html += '  <li><a href="#" onClick="displayNetworks(); return false;" title="Edit &lt;networks&gt; section">Networks</a> - Specify the hosts and network ranges you want to treat specially in your rules.</li>\n';
  html += '  <li><a href="#" onClick="displayReserved(); return false;" title="Edit &lt;reserved&gt; section">Reserved</a> - Specify those network ranges/hosts that are not allowed to come from the External Zone (ie the Internet).</li>\n';
  html += '  <li><a href="#" onClick="displayZones(); return false;" title="Edit &lt;zones&gt; section">Zones</a> - Specify the computer interfaces that exist in each Zone.  Your rules will depend on these settings to limit traffic in, out and through your firewall.</li>\n';
  html += '  <li><a href="#" onClick="displayServices(); return false;" title="Edit &lt;services&gt; section">Services</a> - Specify the services that you are going to reference in your rules.</li>';
  html += '  <li><a href="#" onClick="displayServiceGroups(); return false;" title="Edit &lt;paths&gt; section">Paths</a> - Here you specify the firewall rules that allow traffic in/out/through your firewall.</li>';
  html += '  <li><a href="#" onClick="displayXML(); return false;" title="Display the XML data">Display XML</a> - Display what will be sent to the server.  This is mainly for debugging purposes, but can also be educational.</li>\n';
  html += '  <li>Finally, <b>Save</b> your work! &nbsp;Checking the <b>Backup</b> option allows you to revert to a previous xml config file, if needed. &nbsp;If you want to return to the <b>Main Menu</b> after saving, just select the <b>Save</b> radio button. &nbsp;If you want to Save and then Generate the firewall shell script, select the <b>Generate</b> radio button.</li>\n';
  html += '</ol>\n';
  html += '<br>\n';
  html += '<b>Mini FAQ</b>:<br>\n';
  html += '<table border="0" cellpadding="0" cellspacing="0">\n';
  html += '  <tr>\n';
  html += '    <td><b>Q</b>.</td>\n';
  html += '    <td>What does the VeryTight2 Rule module do?</td>\n';
  html += '  </tr>\n';
  html += '  <tr>\n';
  html += '    <td valign="top"><b>A</b>.</td>\n';
  html += '    <td>The PCXFirewall VeryTight2 Rules module implements a locked down firewall by default. &nbsp;The only traffic you can currently not\n';
  html += '    firewall directly is the local connections. &nbsp;By default the machine can talk to itself on any port.<br><br>\n';
  html += '    All other traffic in, out and through the firewall is blocked unless you implement a rule to allow it in, out or through.<br><br>\n';
  html += '    There is also built-in support to make using <a href="http://www.freeswan.org/" target="links">FreeS/wan VPN\'s</a> extremely easy (from the firewall side, you still have to properly configure your freeswan software - /etc/ipsec.conf, /etc/ipsec.secrets, etc.),\n';
  html += '    working in <a href="http://bridge.sf.net/" target="links">bridged mode</a> and now supporting <a href="http://www.snort.org/" target="links">snort-inline</a>!\n';
  html += '    </td>\n';
  html += '  </tr>\n';
  html += '  <tr><td colspan="2">&nbsp;</td></tr>\n';
  html += '  <tr>\n';
  html += '    <td><b>Q</b>.</td>\n';
  html += '    <td>What are the available Zones?</td>\n';
  html += '  </tr>\n';
  html += '  <tr>\n';
  html += '    <td valign="top"><b>A</b>.</td>\n';
  html += '    <td>The PCXFirewall VeryTight2 Rules module supports the following zones (<b>external</b>, <b>internal</b>, <b>ipsec</b>, <b>dmz</b>, <b>dialin</b>).<br><br>\n';
  html += '    The <b>external</b> zone represents the Internet or some other untrusted network that you want to limit inward and outward connections to from your other Zones (networks).  The external zone is always enabled.<br><br>\n';
  html += '    The <b>internal</b> zone should represent your intranet and usually should only allow traffic out and not in from other Zones unless absolutely needed.<br><br>\n';
  html += '    The <b>ipsec</b> zone was designed to represent <a href="http://www.freeswan.org/" target="links">FreeS/wan</a> IPSec interfaces and allows for filtering your traffic in and out of your VPN connections.<br><br>\n';
  html += '    The <b>dmz</b> zone is used for holding machines you want visible to the World at large, but need to be seperated from your internal network for security reasons.<br><br>\n';
  html += '    The <b>dialin</b> zone allows you to treat modems (ppp0, etc.) as seperate from other interfaces on your firewall and thus implement a more diverse security plan.\n';
  html += '    </td>\n';
  html += '  </tr>\n';
  html += '</table>\n';

  content.innerHTML = html;
}

// message = string to display.
// type = 'info' or 'error'.
function displayStatus(message, type)
{
  var status = getStatus();

  var output = (type == 'info' ? '<b>Status</b>:&nbsp;' : '<b><span style="color: red;">Error</span></b>:&nbsp;') + message;

  status.innerHTML = output;
}

function trueFalseSelectOptions()
{
  var output = '    <option value="1">true</option>\n';
  output += '    <option value="0">false</option>\n';

  return output;
}

function dynamicInterfacesModeSelectOptions()
{
  var output = '    <option value="lookupIP">lookupIP</option>\n';
  output += '    <option value="ignoreIP">ignoreIP</option>\n';

  return output;
}

function selectOutput(name, multi, options)
{
  var output = '  <select name="' + name + '"';
  if (multi)
  {
    output += ' multiple="1"';
  }
  output += '>\n';
  output += options;
  output += '  </select>\n';
}

// section: config, networks, reserved, zones, services, paths
// name: name of value in the section (same as dataObj value)
// displayString: the version of the name to display to the user
// type: text - text field, select - select box, multiselect - multi select box,
//       mixed - text field / select box where text field is name+Value and select is name+Option,
//       bool - checkbox to represent true/false, checkbox - checkbox input field,
//       textarea - textarea input field.
// options: mainly used for select and mixed types, it provides the select options to display
// helpString: the info to display to the user when they click on the displayString help link
// handler: (optional) specify a JS snippet to output in the generated form item.
//          if specified and type = mixed, then the handler is output on the text field.
function createLookupEntry(section, name, displayString, type, options, helpString, handler)
{
  if (arguments.length != createLookupEntry.length && arguments.length < createLookupEntry.length - 1)
  {
    throw new Error("createLookupEntry():  Not enough arguments!  Expected: " + (createLookupEntry.length - 1) + "; Got: " + arguments.length);
  }
  var result = section.match(/^(config|networks|reserved|zones|services|paths|serviceGroup)$/);
  if (result == null)
  {
    throw new Error("createLookupEntry():  Section = '" + section + "' is unknown!");
  }
  result = type.match(/^(text|select|select\+clear|multiselect|mixed|textarea|bool|checkbox)$/);
  if (result == null)
  {
    throw new Error("createLookupEntry():  Type = '" + type + "' is unknown!");
  }

  nameLookups[section][name] = new Object;
  nameLookups[section][name].displayString = displayString;
  nameLookups[section][name].helpString = helpString;
  nameLookups[section][name].type = type;
  nameLookups[section][name].options = options;
  if (arguments.length == createLookupEntry.length)
  {
    nameLookups[section][name].handler = handler;
  }
  else
  {
    nameLookups[section][name].handler = '';
  }
}

function getLookupEntry(section, name)
{
  if (nameLookups[section] == null)
  {
    throw new Error("getLookupEntry(): Section = '" + section + "' is unknown!");
  }
  if (nameLookups[section][name] == null)
  {
    throw new Error("getLookupEntry(): Name = '" + name + "' is unknown!");
  }
  var info = nameLookups[section][name];

  return info;
}

function displayHelp(section, name)
{
  if (nameLookups[section] == null)
  {
    alert("Error displaying Help!  Section = '" + section + "' is unknown!");
    return;
  }
  if (nameLookups[section][name] == null)
  {
    alert("Error displaying Help!  Name = '" + name + "' is unknown!");
    return;
  }
  var info = nameLookups[section][name];
  var w = window.open("", 'helpWindow', 'width=400,height=400');

  w.document.writeln('<html>');
  w.document.writeln('  <head><title>Help</title></head>');
  w.document.writeln('  <body>');
  w.document.writeln('    <center><h1>Help for <i>' + info.displayString + '</i></h1></center>');
  w.document.writeln('    <br><br>' + info.helpString);
  w.document.writeln('    <br><br><a href="#" onClick="window.close(); return false;">Close Window</a>');
  w.document.writeln('  </body>');
  w.document.writeln('</html>');
  w.document.close();
}

function createConntrackerEntry(active, params)
{
  this.active = active;
  this.params = params;
}

function createNetworkEntry(host, type, used, active, comment)
{
  this.host = host;
  this.type = type;
  this.used = used;
  this.active = active;
  this.comment = comment;
}

function createReservedEntry(host, limitTo, active, comment)
{
  //alert("createReservedEntry limitTo = '" + limitTo.toString() + "'.");
  this.host = host;
  this.limitTo = limitTo;
  this.active = active;
  this.comment = comment;
}

function createIPSecZoneEntry(name, boundTo, network, proxyarp, active, comment)
{
  this.name = name;
  this.network = network;
  this.proxyarp = proxyarp;
  this.active = active;
  this.comment = comment;
  this.boundTo = boundTo;
}

function createOtherZoneEntry(name, ip, ipsec, ipsecLimitFrom, network, proxyarp, active, comment)
{
  this.name = name;
  this.network = network;
  this.proxyarp = proxyarp;
  this.active = active;
  this.comment = comment;
  this.ip = ip;
  this.ipsec = ipsec;
  this.ipsecLimitFrom = ipsecLimitFrom;
}

function createServiceInfoEntry(proto, dport, sport, icmpType)
{
  this.proto = proto;
  this.dport = dport;
  this.sport = sport;
  this.icmpType = icmpType;
}

function createServiceGroupInfoEntry(i, ip, o, limitFrom, limitTo, reject, dropNoLog, dnat, dnatDestination, dnatZone, dnatInterface, dnatPort, redirect, redirectPort, queue, log, logComment, logLevel, service, active, comment)
{
  this.i = i;
  this.ip = ip;
  this.o = o;
  this.limitFrom = limitFrom;
  this.limitTo = limitTo;
  this.reject = reject;
  this.dropNoLog = dropNoLog;
  this.dnat = dnat;
  this.dnatDestination = dnatDestination;
  this.dnatZone = dnatZone;
  this.dnatInterface = dnatInterface;
  this.dnatPort = dnatPort;
  this.redirect = redirect;
  this.redirectPort = redirectPort;
  this.queue = queue;
  this.log = log;
  this.logComment = logComment;
  this.logLevel = logLevel;
  this.service = service;
  this.active = active;
  this.comment = comment;
}

function isTrueSelect(index)
{
  if (index == 0)
  {
    return 1;  // true
  }
  else
  {
    return 0;  // false
  }
}

function boolSelectIndex(value)
{
  if (value == 1)
  {
    return 0;  // true
  }
  else
  {
    return 1;  // false
  }
}

function dynamicInterfacesSelectIndex(value)
{
  if (value == "lookupIP")
  {
    return 0;
  }
  else
  {
    return 1;
  }
}

function selectAllButton(item)
{
  // return the code to display a Select All button that works with form item named item.
  var output = '';

  output += '<input type="button" name="selectAll' + item + '" value="Select All" onClick="selectAll(\'' + item + '\'); entryChangesMade=true">';

  return output;
}

function toggleSelectButton(item)
{
  // return the code to display a Toggle button that works with form item named item.
  var output = '';

  output += '<input type="button" name="toggleSelect' + item + '" value="Toggle" onClick="toggleSelection(\'' + item + '\'); entryChangesMade=true;">';

  return output;
}

// enable or disable the selectAll and toggleSelect buttons named item.
function displaySelectButtons(item, state)
{
  displaySelectAllButton(item, state);
  displayToggleSelectButton(item, state);
}

// enable or disable the selectAll button named item.
function displaySelectAllButton(item, state)
{
  var config = document.configF;

  eval('config.selectAll' + item + '.disabled=' + state);
}

// enable or disable the toggleSelect button named item.
function displayToggleSelectButton(item, state)
{
  var config = document.configF;

  eval('config.toggleSelect' + item + '.disabled=' + state);
}

function selectAll(item)
{
  var config = document.configF;

  eval('var itemF = config.' + item);

  if (itemF.type != "select-multiple")
  {
    alert("Form item = '" + item + "' is not a multi-select select box!");
  }
  else
  {
    for (var i=0; i < itemF.options.length; i++)
    {
      itemF.options[i].selected = true;
    }
  }
}

function toggleSelection(item)
{
  var config = document.configF;

  eval('var itemF = config.' + item);

  if (itemF.type != "select-multiple")
  {
    alert("Form item = '" + item + "' is not a multi-select select box!");
  }
  else
  {
    for (var i=0; i < itemF.options.length; i++)
    {
      itemF.options[i].selected = !itemF.options[i].selected;
    }
  }
}

function numHashEntries(array)
{
  // figure out how many entries are in the array and return the count.
  var count = 0;

  for (var i in array)
  {
    count++;
  }

  return count;
}

// recursively copy the contents of the array, cloning any arrays of arrays, etc.
function cloneArray(array)
{
  var newArray = new Array();

  for (var i in array)
  {
    newArray[i] = ((typeof array[i] == "object") ? ((array[i] instanceof Array) ? cloneArray(array[i]) : array[i]) : array[i]);
  }

  return newArray;
}

// fixup the input to escape <> in the output string so we prevent XSS exploits.
function fixupOutput(output)
{
  var output2 = output.replace(/</g, "&lt;");
  output2 = output2.replace(/>/g, "&gt;");

  return output2;
}
